/*************************************************************
Generated for IBM Rational SDL & TTCN Suite 6.3.0.2.22515 at Mon Nov 21 06:47:45 PM CET 2011
Licensed Materials - Property of IBM 
(c) Copyright IBM Corporation 1990, 2011.  All Rights Reserved.

U.S. Government Users Restricted Rights:  
Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*************************************************************/
/*+MHDR*/
/*
+------------------------------------------------------------------------------+
#  Modulname: mt_cod.c
+------------------------------------------------------------------------------+
|                                                                              |
|  Description  :                                                              |
|    This module contains the coding and decoding of information, which is     |
|    exchanged via the communication interface.                                |
|                                                                              |
|    The functions exported by this module are necessary for the target only.  |
|                                                                              |
|    The following problems are solved by the functions of this module :       |
|      * defining a unique protocol between two communicating processorsystems |
|      * function(s) to encode C structures into a protocol data unit.         |
|      * function(s) to decode a protocol data unit into a C structure.        |
|      * allow all entities in the system to use the communications link.      |
|      * all encoding / decoding is performed within this module, except the   |
|        encoding / decoding of user messages. users are free to use the       |
|        functions of the DL- module.                                          |
|                                                                              |
|    The intelligence to encode / decode informations resides                  |
|    on the host. The target only copies the data which must be sent/received, |
|    and does not know about types of SDL data.                                |
|                                                                              |
|    The following problems are NOT SOLVED by the functions of this module :   |
|      * protocol for the parameters of SDL Signals, which are exchanged via   |
|        the communication interface.                                          |
|                                                                              |
|    See the Cmicro Protocol definition below.                                 |
|                                                                              |
|                                                                              |
|  M1 Errormessages, which are routed to the ErrorHandler                      |
|  --------------------------------------------------------                    |
|    ERR_N_LINK_NOT_IMPL         Errorcheck is conditionally compiled          |
|                                      If a feature is not yet implemented.    |
|    ERR_N_LINK_SYNC             Errorcheck is conditionally compiled          |
|                                      If the synchronization is lost          |
|                                      within a frame of the Cmicro Protocol.  |
|    ERR_N_LINK_CHKSUM           Errorcheck is conditionally compiled          |
|                                      If the checksum of the CRC has been     |
|                                      failed in a received frame of the Cmicro|
|                                      Protocol.                               |
|    ERR_N_LINK_DEST_BUFFER      Errorcheck is conditionally compiled          |
|                                      If the message, which is to be decoded, |
|                                      doesn't fit into the destination buffer.|
|                                      Probably there is an inconsistancy in   |
|                                      the received length information !       |
|    ERR_N_LINK_ENC_MEMORY       Errorcheck is conditionally compiled          |
|                                      If is not possible to allocate one      |
|                                      more buffer, which is necessary to en-  |
|                                      code a frame of the Cmicro Protocol.    |
|    ERR_N_LINK_ENC_LENGTH       Errorcheck is conditionally compiled          |
|                                      If the encode mechanism fails due to    |
|                                      an illegal length.                      |
|    ERR_N_DATA_LINK             Errorcheck is conditionally compiled          |
|                                      Data link doesn't respond (cannot       |
|                                      send encoded frame)                     |
|    ERR_N_TESTER_MESSAGE        Errorcheck is conditionally compiled          |
|                                      Message was treated as a Cmicro Tester  |
|                                      Message, but cannot be decoded.         |
|                                                                              |
|    ERR_N_DECODE_METHOD         Errorcheck is conditionally compiled          |
|                                      There is no defined method to decode    |
|                                      the received frame.                     |
|                                                                              |
|                                                                              |
|  M2 Exported functions of this module :                                      |
|  --------------------------------------------------------                    |
|      int  [Qualified]GetHeaderLength     ( )                                 |
|      int  [Qualified]DecodeHeader        ( )                                 |
|      int  [Qualified]Cod_Decode          ( )                                 |
|      int  [Qualified]Cod_Encode          ( )                                 |
|      char [Qualified]Cod_CheckSum        ( )                                 |
|                                                                              |
|  M3 Static functions of this module :                                        |
|  --------------------------------------------------------                    |
|      -                                                                       |
|                                                                              |
|  The definition of the Cmicro Protocol :                                     |
|  =======================================                                     |
|                                                                              |
|  Each Message consists of :                                                  |
|                                                                              |
|        tag1 tag2 SYNC length data CRC                                        |
|                                                                              |
|  where :                                                                     |
|        SYNC = 0xaa                                                           |
|        CRC  = see Data Link Module                                           |
|        length = sizeof (data) without CRC                                    |
|                                                                              |
|    tag1 :       tag1 :  Comment                                              |
|    ------------------------------------------------------------------------- |
|    0000 0000 B  0x00  : free for the user of the Cmicro Protocol             |
|       ....      ...                  ......                                  |
|    1110 1111 B  0xef  : free for the user of the Cmicro Protocol             |
|                                                                              |
|    1111 0000 B  0xf0  : message for XMK_MICRO_TRACER                         |
|                         tag2 shows message type / tag3 shows length          |
|                         if bit 8 == set in tag2, then 2 bytes length         |
|                         else 1 byte length                                   |
|                         The first byte which is transmitted for length       |
|                         is always the higher byte, followed by the lower one.|
|                                                                              |
|    1111 0001 B  0xf1  : message for XMK_MICRO_COMMAND                        |
|                         tag2 shows message type / tag3 shows length          |
|                         if bit 8 == set in tag2, then 2 bytes length         |
|                         else 1 byte length                                   |
|                         The first byte which is transmitted for length       |
|                         is always the higher byte, followed by the lower one.|
|                                                                              |
|    1111 0010 B  0xf2  : message for XMK_MICRO_RECORDER                       |
|                         tag2 shows message type / tag3 shows length          |
|                         if bit 8 == set in tag2, then 2 bytes length         |
|                         else 1 byte length                                   |
|                         The first byte which is transmitted for length       |
|                         is always the higher byte, followed by the lower one.|
|                                                                              |
|    1111 0011 B  0xf3  : reserved for internal purposes                       |
|       ....       ...                                                         |
|    1111 1111 B  0xff  : reserved for internal purposes                       |
|                                                                              |
|    Coding of tag 2    :                                                      |
|    ------------------------------------------------------------------------- |
|    The interpretation of tag 2 depends on the value of tag 1. If tag 1       |
|    is one of the values for XMK_MICRO_*, then tag 2 is coded as one of       |
|    the values from CMD_T* (i.e.CMD_TTASK for SDL Task).                      |
|                                                                              |
|    If Bit 7 of tag 2 is set, then the length is coded as 2 Bytes length.     |
|                                                                              |
|    Coding of SYNC     : 0xaa                                                 |
|    ------------------------------------------------------------------------- |
|                                                                              |
|    Coding of CRC                                                             |
|    ------------------------------------------------------------------------- |
|    The CRC is a value of type unsigned char. It is evaluated for the         |
|    whole buffer by adding with modulo 256.                                   |
|                                                                              |
+------------------------------------------------------------------------------+
*/
/*-MHDR*/

#ifndef __MT_COD_C_
#define __MT_COD_C_

/*+IMPORT*/
/*====================  I M P O R T  =========================================*/
#include "ml_typ.h"

/*============================================================================*/
/*-IMPORT*/


/*+MGG*/
/*====================  V A L U E S  O F  T H I S   M O D U L E  =============*/


/*--------------------  Constants,  Macros  ----------------------------------*/

/*--------------------  Typedefinitions     ----------------------------------*/

/*--------------------  Functions    -----------------------------------------*/

/*--------------------  Variables    -----------------------------------------*/
#ifdef XMK_ADD_MICRO_TESTER
  static xmk_T_BOOL CRC_Comparison_wanted = XMK_FALSE;
#endif

/*============================================================================*/
/*-MGG*/

#ifdef XMK_ADD_MICRO_TESTER

#ifdef XMK_ADD_MICRO_RECORDER

/*+FHDR E*/
/*
+------------------------------------------------------------------------------+
|  Functionname : [Qualified]GetHeaderLength  - compiled for target only       |
+------------------------------------------------------------------------------+
|                                                                              |
|  Description :                                                               |
|    This function decodes the informations header, which have been received on|
|    the communications interface.                                             |
|                                                                              |
|  Parameter    : char *p_RAWBuffer     IN   Pointer to the RAW buffer, which  |
|                                            is to be analyzed.                |
|                                                                              |
|  Return       : 1 - x                 OUT  Amount of headerbytes including   |
|                                            the length field.                 |
|                                                                              |
|                                                                              |
+------------------------------------------------------------------------------+
*/
/*-FHDR E*/

/*+FDEF E*/
int
xmk_GetHeaderLength (
  char* p_RAWBuffer )
/*-FDEF E*/
{
  /*
  ** suppress compiler warning
  */
  XMK_SUPPRESS(p_RAWBuffer);

  return (4 * XMK_SIZEOF_CHAR) ;
}
#endif /* ... XMK_ADD_MICRO_RECORDER */


/*+FHDR E*/
/*
+------------------------------------------------------------------------------+
|  Functionname : [Qualified]DecodeHeader      - compiled for target only      |
+------------------------------------------------------------------------------+
|                                                                              |
|  Description :                                                               |
|    This function decodes the informations header, which have been received on|
|    the communications interface.                                             |
|                                                                              |
|    Basically, a mem copy is used to transfer all informations from           |
|    and to the the communications interface.                                  |
|                                                                              |
|    If the header is incomplete (i.e.not enough bytes received via the        |
|    comm.interface), the function returns with XMK_WAIT.                      |
|                                                                              |
|    If it returns with XMK_OKAY, then the OUT values are set correctly.       |
|                                                                              |
|  Parameter    : char*p_data           IN   pointer to hexbuffer. The pointer |
|                                            size.                             |
|                 int * p_MessageClass  OUT  returns the message class         |
|                 int * p_MessageTag    OUT  returns the message tag           |
|                 int * p_MessageLength OUT  returns the message length        |
|                                            without header == net data        |
|                                                                              |
|  Return       : XMK_OKAY              OUT  Okay                              |
|                 XMK_WAIT                   Incomplete Header                 |
|                 XMK_ERROR                  Anything went wrong               |
|                                                                              |
+------------------------------------------------------------------------------+
*/
/*-FHDR E*/

/*+FDEF E*/
int
xmk_DecodeHeader (
  char * p_RAWBuffer,
  int    RAWBufferLength,
  int  * p_MessageClass,
  int  * p_MessageTag,
  int  * p_MessageLength )
/*-FDEF E*/
{
  char  length2 = XMK_FALSE ; /* 2 bytes length to encode or not ? */
  int   length  = 0         ;
  unsigned char  sync_byte = 0 ;
  unsigned char  header;

  #ifdef XMK_ADD_PRINTF_COD
    XMK_FUNCTION("[Qualified]DecodeHeader");
  #endif

  /*
  ** Minimum Header complete ?
  */
  header = 4 * XMK_SIZEOF_CHAR;

  if( RAWBufferLength < header )
    return( XMK_WAIT );

  /*
  ** ---------------------------------------------------
  ** Decoding the Message Class ...
  ** ---------------------------------------------------
  */
  *p_MessageClass = (int) (unsigned char) *p_RAWBuffer;
  p_RAWBuffer ++;
  *p_MessageClass = (*p_MessageClass & 0xff) ;

  /*
  ** ---------------------------------------------------
  ** Decoding the Message Tag ...
  ** ---------------------------------------------------
  */
  *p_MessageTag = (int) *p_RAWBuffer;
  p_RAWBuffer ++;
  *p_MessageTag = (*p_MessageTag & 0xff) ;

  if( *p_MessageTag & 0x80)
  {
    /*
    ** 2 octets length field
    */
    *p_MessageTag = (*p_MessageTag & 0x007f);
    length2 =XMK_TRUE;
    if( RAWBufferLength < ( header + 1 ) )
      return( XMK_WAIT );
  }


  /*
  ** ---------------------------------------------------
  ** Decode and check the SYNC Byte ...
  ** ---------------------------------------------------
  */
  sync_byte = (unsigned char) *p_RAWBuffer;
  p_RAWBuffer ++;
  if (sync_byte != SYNC_BYTE)
  {
    return (XMK_ERROR);
  }



  /*
  ** ---------------------------------------------------
  ** Decoding the Message Length ...
  ** ---------------------------------------------------
  */
  if( length2 == XMK_TRUE )
  {
    length = (int) (unsigned char)(*p_RAWBuffer & 0xff );
    length <<= 8;
    p_RAWBuffer ++;
    length += (int) (unsigned char)(*p_RAWBuffer & 0xff );
    *p_MessageLength = length;
  }
  else
  {
    *p_MessageLength = (int)(unsigned char) *p_RAWBuffer;
  }
  p_RAWBuffer ++;

#ifdef XMK_ADD_PRINTF
  fprintf (stdout, "DataLength(words): %d\n",  length);
  fprintf (stdout, "DataLength(bytes): %d\n",  (length * XMK_SIZEOF_CHAR) );
  fprintf (stdout, "BufferLength     : %d\n",  *p_MessageLength);
#endif

  return (XMK_OKAY);
}

/*+FHDR E*/
/*
+------------------------------------------------------------------------------+
|  Functionname : [Qualified]Cod_Decode     - compiled for target only         |
+------------------------------------------------------------------------------+
|                                                                              |
|  Description :                                                               |
|    This function decodes the information, which has been received on         |
|    the communications interface, into C structures.                          |
|                                                                              |
|  Parameter    : char*p_RAWBuffer      IN   pointer to hexbuffer. The pointer |
|                                            must point to tag 1 (the message- |
|                                            class), if this is used, or to    |
|                                            tag2 (message tag) if message-    |
|                                            Class is not used.                |
|                 int * p_MessageClass  OUT  returns the message class         |
|                 int * p_MessageTag    OUT  returns the message tag           |
|                 int * p_MessageLength OUT  returns the message length in sum |
|                 char*p_dest           OUT  pointer to appropriate C struc-   |
|                                            ture. The caller has to ensure,   |
|                                            that the buffer, where this       |
|                                            pointer points to, has the correct|
|                                            size.                             |
|                 int    dest_len            size of destination buffer        |
|                                                                              |
|  Return       : XMK_OKAY              OUT  Okay                              |
|                 XMK_ERROR                  Anything went wrong               |
|                 XMK_IGNORE                 Data not for Cmicro Tester        |
|                                            (xInEnvHost () has been called)   |
|                                                                              |
+------------------------------------------------------------------------------+
*/
/*-FHDR E*/

/*+FDEF E*/
int
xmk_Cod_Decode (
  char * p_RAWBuffer,
  int  * p_MessageClass,
  int  * p_MessageTag,
  int  * p_MessageLength,
  char * p_dest,
  int    dest_len )
/*-FDEF E*/
{
  int        length  = 0         ;
  char       length2 = XMK_FALSE ; /* 2 bytes length to encode or not ? */
  int        cnt     = 0         ;
  unsigned   char  sync_byte     ;
  char       evaluated_chksum    ;
  char       received_chksum     ;
  char      *p_read              ;

  #ifdef XMK_ADD_PRINTF_COD
    XMK_FUNCTION("[Qualified]Cod_Decode");
  #endif


  /*
  ** Initialization ...
  */
  p_read = p_RAWBuffer;

  /*
  ** ---------------------------------------------------
  ** Disabling other processes from execution ...
  ** ---------------------------------------------------
  */
  XMK_BEGIN_CRITICAL_PATH ;

  /*
  ** ---------------------------------------------------
  ** Decoding the Message Class ...
  ** ---------------------------------------------------
  */
  *p_MessageClass  = (int) ( (*p_read) & 0xff);
  p_read++; cnt += XMK_SIZEOF_CHAR;
  *p_MessageClass = (*p_MessageClass & 0xff) ;

  /*
  ** ---------------------------------------------------
  ** Decoding the Message Tag ...
  ** ---------------------------------------------------
  */
  *p_MessageTag = (int) ( (*p_read) & 0xff);
  p_read++; cnt += XMK_SIZEOF_CHAR;

  if (*p_MessageTag & 0x80)
  {
    length2 = XMK_TRUE    ;
  }

  /*
  ** ---------------------------------------------------
  ** Decode and check the SYNC Byte ...
  ** ---------------------------------------------------
  */
  sync_byte = (unsigned char) (*p_read & 0xff);
  p_read++; cnt += XMK_SIZEOF_CHAR;
  if (sync_byte != SYNC_BYTE)
  {
    XMK_END_CRITICAL_PATH ;
    ErrorHandler (ERR_N_LINK_SYNC);
    return (XMK_ERROR);
  }

  /*
  ** ---------------------------------------------------
  ** Decoding the Message Length ...
  ** ---------------------------------------------------
  */
  if( length2 == XMK_TRUE )
  {
    length = (int) ( (*p_read) & 0xff);
    length <<= 8;
    p_read++; cnt += XMK_SIZEOF_CHAR;
    length += (int) ( (*p_read) & 0xff);
    *p_MessageLength = length;
  }
  else
  {
    *p_MessageLength = (int) ( (*p_read) & 0xff);
  }
  p_read++; cnt += XMK_SIZEOF_CHAR;

  /*
  ** ---------------------------------------------------
  ** Decode the checksum
  ** ---------------------------------------------------
  */
  received_chksum  = * ( p_read + *p_MessageLength) ;

  /*
  ** ---------------------------------------------------
  ** Evaluate the checksum
  ** ---------------------------------------------------
  */
  evaluated_chksum = xmk_Cod_CheckSum ( (char*) p_RAWBuffer,
                                         cnt + *p_MessageLength );

  /*
  ** ---------------------------------------------------
  ** Compare the checksum
  ** ---------------------------------------------------
  */
  if (CRC_Comparison_wanted && (evaluated_chksum != received_chksum))
  {
    XMK_END_CRITICAL_PATH ;
    ErrorHandler (ERR_N_LINK_CHKSUM);
    return (XMK_ERROR);
  }

  /*
  ** ------------------------------------------------------------
  ** Errorcheck : Does Message fit into destination buffer ?
  ** ------------------------------------------------------------
  */
  if (dest_len < *p_MessageLength)
  {
    XMK_END_CRITICAL_PATH ;
    ErrorHandler (ERR_N_LINK_DEST_BUFFER);
    return (XMK_ERROR)    ;
  }

  /*
  ** ---------------------------------------------------
  ** Decoding the Message data ...
  ** ---------------------------------------------------
  */
  memcpy ( p_dest, p_read, *p_MessageLength );

  /*
  ** -------------------------------------------------------------
  ** Allow other processes to be executed ...
  ** -------------------------------------------------------------
  */
  XMK_END_CRITICAL_PATH ;
  return ( XMK_OKAY );
}

/*+FHDR E*/
/*
+------------------------------------------------------------------------------+
|  Functionname : [Qualified]Cod_Encode     - compiled for target only         |
+------------------------------------------------------------------------------+
|                                                                              |
|  Description :                                                               |
|    This function encodes the given data according to the following           |
|    scheme :                                                                  |
|                                                                              |
|    tag1 tag2 length [length2]                                                |
|                                                                              |
|    See the protocol definition in Module header.                             |
|                                                                              |
|  Parameter    : int MessageClass                                             |
|                 int MessageTag                                               |
|                 char *p_data                                                 |
|                 int length                                                   |
|                                                                              |
|  Return       : XMK_OKAY                                                     |
|                 XMK_ERROR    cannot encode, due to "no free avail.memory",   |
|                              error in data link, etc....                     |
|                                                                              |
+------------------------------------------------------------------------------+
*/
/*-FHDR E*/

/*+FDEF E*/
int
xmk_Cod_Encode (
  int MessageClass,
  int MessageTag,
  char *p_data,
  int length )
/*-FDEF E*/
{
  int   write_result        ;
  int   result = XMK_OKAY   ;
  int   cnt    = 0          ;
  char  xmk_RAM_ptr pbuf    ;
  char  xmk_RAM_ptr pwrite  ;
  char  length2 = XMK_FALSE ; /* 2 bytes length to encode or not ? */

  #ifdef XMK_ADD_PRINTF_COD
    XMK_FUNCTION("[Qualified]Cod_Encode");
  #endif

  /*
  ** ---------------------------------------------------
  ** Disabling other processes from execution ...
  ** ---------------------------------------------------
  */
  XMK_BEGIN_CRITICAL_PATH ;

  /*
  ** ---------------------------------------------------
  ** Buffer allocation ...
  ** Space must be left for the header and CRC !!
  ** ---------------------------------------------------
  */
  pbuf = (char xmk_RAM_ptr) xAlloc ((length * XMK_SIZEOF_CHAR) + 20 );
  if (pbuf == (char xmk_RAM_ptr) NULL)
  {
    XMK_END_CRITICAL_PATH ;
    ErrorHandler (ERR_N_LINK_ENC_MEMORY);
    return (XMK_ERROR);
  }

  /*
  ** ---------------------------------------------------
  ** Buffer initialization ...
  ** ---------------------------------------------------
  */
  pwrite = pbuf;

  /*
  ** ---------------------------------------------------
  ** Encode the Message Class
  ** ---------------------------------------------------
  */
  *pwrite = (unsigned char) MessageClass;
  pwrite++; cnt++;

  /*
  ** ---------------------------------------------------
  ** Encode the Message Tag
  ** ---------------------------------------------------
  */
  *pwrite = (unsigned char) MessageTag ;

  if (length >= 255)
  {
    length2 = XMK_TRUE;
    *pwrite = (*pwrite | 0x80);
  }

  pwrite++; cnt++;

  /*
  ** ---------------------------------------------------
  ** Encode the SYNC Byte
  ** ---------------------------------------------------
  */
  *pwrite = (unsigned char) SYNC_BYTE ;
  pwrite++; cnt++;


  /*
  ** ---------------------------------------------------
  ** Encode the Message Length
  ** ---------------------------------------------------
  */
  if (length2 == XMK_TRUE)
  {
    *pwrite = (unsigned char) (length >> 8);
    pwrite++; cnt++;
  }

  /*
  ** the lower octet of the length has to be used in any case
  */
  *pwrite = (unsigned char) (length & 0xff) ;
  pwrite++; cnt++;

  /*
  ** -------------------------------------------------
  ** Encode the data ....
  ** -------------------------------------------------
  */
  if (length != 0)
  {
    memcpy (pwrite, p_data, length);
    cnt    += length    ;
    pwrite += length    ;
  }
  result  = XMK_OKAY  ;

  /*
  ** ---------------------------------------------------
  ** Encode the Checksum
  ** ---------------------------------------------------
  */
  *pwrite = (unsigned char) xmk_Cod_CheckSum ( pbuf, cnt ) ;
   pwrite++; cnt++;

  XMK_END_CRITICAL_PATH ;

  /*
  ** -------------------------------------------------------------
  ** Writing the encoded informations to the
  ** serial communications interface
  ** -------------------------------------------------------------
  */

  write_result = xmk_DLWrite (pbuf, cnt) ;
  if (write_result != XMK_OKAY)
  {
    ErrorHandler (ERR_N_DATA_LINK);
    result = XMK_ERROR ;
  }


  /*
  ** -------------------------------------------------------------
  ** Give up the temporary buffer ...
  ** Allow other processes to be executed ...
  ** -------------------------------------------------------------
  */
  xFree ( (void xmk_RAM_ptr xmk_RAM_ptr) & pbuf );
  return (result);
}



/*+FHDR E*/
/*
+------------------------------------------------------------------------------+
|  Functionname : [Qualified]Cod_CheckSum   - evaluate checksum of buffer      |
+------------------------------------------------------------------------------+
|                                                                              |
|  Description :                                                               |
|                                                                              |
|    This function evaluates the checksum of a buffer with a given length      |
|    and returns with the resultvalue.                                         |
|                                                                              |
|    The function is necessary in all configurations.                          |
|                                                                              |
|  Parameter    : char * buffer                                                |
|                 int    buffer_length                                         |
|                                                                              |
|  Return       : unsigned char   checksum wert                                |
|                                                                              |
+------------------------------------------------------------------------------+
*/
/*-FHDR E*/

/*+FDEF E*/
char
xmk_Cod_CheckSum (
  char *p_RAWBUFFER,
  int RAWBufferLength )
/*-FDEF E*/
{
  unsigned char  chksum = 0;

  while (RAWBufferLength > 0)
  {
    chksum          += *p_RAWBUFFER ;
    p_RAWBUFFER     ++ ;
    RAWBufferLength -- ;
  }
  return (chksum);
}

#endif /* ... XMK_ADD_MICRO_TESTER */

#endif /* __ML_COD_C_ */
