/*************************************************************
Generated for IBM Rational SDL & TTCN Suite 6.3.0.2.22515 at Mon Nov 21 06:47:45 PM CET 2011
Licensed Materials - Property of IBM 
(c) Copyright IBM Corporation 1990, 2011.  All Rights Reserved.

U.S. Government Users Restricted Rights:  
Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*************************************************************/


/* 
  Platform-independent socket API for stream sockets
  
  Current support:
    Windows sockets 1.1
    BSD sockets
     - Solaris
     - VxWorks
     - OSE

  Default protocol is TCP/IP

  ---

  The following prefix naming applies: SCM_name

  SCM: Socket Communication Module

  ---

*/

#ifndef XSCT_SOCK
#define XSCT_SOCK


#ifdef THREADWIN32
  /* Windows */

#include "windows.h"
#include "winsock.h"

#define SCM_QUEUE_LENGTH        SOMAXCONN

#define SCM_SHUT_SENDRECV       2

#define SCM_STARTUP             WSADATA wsaData;\
                                int xRetCode;\
                                if (!xSCMInitialized)\
                                {\
                                  xRetCode = WSAStartup(MAKEWORD(1,1), &wsaData);\
                                  if (xRetCode != 0) { SCM_ERROR_REPORT("ERROR WSAStartup: %d\n", xRetCode) return;}\
                                  xSCMInitialized = TRUE;\
                                }

/* host address data structure */
#define SCM_HOST_ENTRY          struct hostent*

#define SCM_SOCKET_ERROR        SOCKET_ERROR
#define SCM_GET_ERROR           WSAGetLastError()


/* Socket creation error */
#define SCM_INVALID_SOCKET      INVALID_SOCKET


#elif THREADVXWORKS
/* VxWorks */

#include "sockLib.h"
#include "hostLib.h"
#include "errno.h"

#define SCM_STARTUP               

#define SOCKET                    int

#define SCM_SOCK_DATA             int

#define SCM_HOST_ENTRY            int

#define SCM_SHUT_SENDRECV         2

#define SCM_GET_ERROR             errno



/* Socket creation error */
#define SCM_INVALID_SOCKET        ERROR

/* General error code */
#define SCM_SOCKET_ERROR          ERROR

/* Queue length */
#define SCM_QUEUE_LENGTH          3 /* can be changed if desired */


#elif THREADOSE
/* OSE */

#include "inet.h"
#include "ineterr.h"
#include "nameser.h"
#include "resolv.h"
#include "netdb.h"

/* The inet and resolv threads are created dynamically */
#define SCM_STARTUP        /*       SCM_THREAD xResolvThr; \
                                  SCM_THREAD xInetThr; \
                                  SCM_THREAD xInitThr; \
SCM_START_THREAD(xInetThr, ose_inet_, 0, 1024, 10) \
                                  SCM_START_THREAD(xInitThr, init_inet_, 0, 256, 9) \
                                  SCM_START_THREAD(xResolvThr, ose_resolve_, 0, 1000, 12)*/

#define SOCKET                    int

#define SCM_SOCK_DATA             int

#define SCM_HOST_ENTRY            struct hostent*

#define SCM_SHUT_SENDRECV         2

#define SCM_GET_ERROR             errno


/* Socket creation error */
#define SCM_INVALID_SOCKET        -1

/* General error code */
#define SCM_SOCKET_ERROR          -1

/* Queue length */
#define SCM_QUEUE_LENGTH          3 /* can be changed if desired */


#elif THREADSOLARIS
/* SOLARIS */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "unistd.h"
#include <errno.h>

#define SCM_STARTUP               

#define SOCKET                    int

#define SCM_SOCK_DATA             int

#define SCM_HOST_ENTRY            struct hostent*

#define SCM_SHUT_SENDRECV         SHUT_RDWR

/* Error handling */
#define SCM_GET_ERROR             errno


/* Socket creation error */
#define SCM_INVALID_SOCKET        -1

/* General error code */
#define SCM_SOCKET_ERROR          -1

/* Queue length */
#define SCM_QUEUE_LENGTH          3 /* can be changed if desired */

#endif /* end platform dependent */


/* receive buffer size */
#define SCM_RECVBUF_MAXSIZE       256

/* socket data type */
#define SCM_SOCKET                SOCKET

/* address data type */
#define SCM_SOCKET_ADDR           struct sockaddr_in

#define SCM_PTR_SOCK_ADDR         (struct sockaddr *)

#define SCM_FDSET                 fd_set

#define SCM_TIMEVAL               struct timeval

/* socket (creates a socket) */
#ifndef SCM_CREATE_SOCKET
  #define SCM_CREATE_SOCKET(DOMAIN, TYPE, PROTOCOL) \
    socket(DOMAIN, TYPE, PROTOCOL)
#endif


/* gethostname (returns name of local host) */
#ifndef SCM_GETHOSTNAME
    #define SCM_GETHOSTNAME(HOST_NAME, NAME_LEN) \
      gethostname(HOST_NAME, NAME_LEN)
#endif


/* gethostbyname (returns info about a host) */
#ifndef SCM_GETHOSTBYNAME
  #ifdef THREADOSE
    #define SCM_GETHOSTBYNAME(HOST_NAME) \
      gethostbyname_r(HOST_NAME)
  #elif THREADVXWORKS
    #define SCM_GETHOSTBYNAME(HOST_NAME) \
      hostGetByName(HOST_NAME)
  #else
    #define SCM_GETHOSTBYNAME(HOST_NAME) \
      gethostbyname(HOST_NAME)
  #endif
#endif


/* bind (connects a socket to an address) */
#ifndef SCM_BIND
  #define SCM_BIND(SOCKET_DESCR, SOCKET_ADDRESS, ADDRESS_LENGTH) \
    bind(SOCKET_DESCR, SOCKET_ADDRESS, ADDRESS_LENGTH)
#endif


/* listen (the server listens for incoming connections) */
/* The queue length is hidden. */
#ifndef LISTEN
  #define SCM_LISTEN(SOCKET_DESCR) \
    listen(SOCKET_DESCR, SCM_QUEUE_LENGTH)
#endif


/* connect (the client attempts to set up a connection) */
#ifndef SCM_CONNECT
  #define SCM_CONNECT(SOCKET_DESCR, SERVER_ADDR, ADDRESS_LENGTH) \
    connect(SOCKET_DESCR, SERVER_ADDR, ADDRESS_LENGTH)
#endif


/* accept (the server accepts a connection attempt from a client) */
#ifndef SCM_ACCEPT
  #define SCM_ACCEPT(SOCKET_DESCR, CLIENT_ADDR, ADDRESS_LENGTH) \
    accept(SOCKET_DESCR, CLIENT_ADDR, ADDRESS_LENGTH)
#endif


/* send (send data on a socket) */
#ifndef SCM_SEND
  #ifdef THREADOSE
    #define SCM_SEND(SOCKET_DESCR, BUFFER, BUFFER_LENGTH, FLAGS) \
      inet_send(SOCKET_DESCR, BUFFER, BUFFER_LENGTH, FLAGS)
  #else
    #define SCM_SEND(SOCKET_DESCR, BUFFER, BUFFER_LENGTH, FLAGS) \
      send(SOCKET_DESCR, BUFFER, BUFFER_LENGTH, FLAGS)
  #endif
#endif
  

/* recv (receives data on a socket) */
#ifndef SCM_RECV
  #define SCM_RECV(SOCKET_DESCR, BUFFER, BUFFER_LENGTH, FLAGS) \
    recv(SOCKET_DESCR, BUFFER, BUFFER_LENGTH, FLAGS)
#endif

/* shutdown */
#ifndef SCM_SHUTDOWN
#define SCM_SHUTDOWN(SOCKET_DESCR, MODE) \
  shutdown(SOCKET_DESCR, MODE)
#endif

/* close (closes a socket) */
#ifndef SCM_CLOSE
  #ifdef THREADWIN32
    #define SCM_CLOSE(SOCKET_DESCR) \
      closesocket(SOCKET_DESCR)
  #else
    #define SCM_CLOSE(SOCKET_DESCR) \
      close(SOCKET_DESCR)
  #endif
#endif


/* select (used on iterative servers to identify events) */
#ifndef SCM_SELECT
  #define SCM_SELECT(NBR_DESCR, READ_ADDR, WRITE_ADDR, EXCEPTION_ADDR, WAIT_TIME) \
    select(NBR_DESCR, READ_ADDR, WRITE_ADDR, EXCEPTION_ADDR, WAIT_TIME)
#endif

#ifndef SCM_FD_SET
#define SCM_FD_SET(S, SET)            FD_SET(S, SET)                     
#endif

#ifndef SCM_FD_ISSET
#define SCM_FD_ISSET(S, SET)          FD_ISSET(S, SET)
#endif

#ifndef SCM_FD_ZERO
#define SCM_FD_ZERO(SET)              FD_ZERO(SET)
#endif

#ifndef SCM_FD_CLR
#define SCM_FD_CLR(S, SET)            FD_CLR(S, SET)
#endif

/* Byte ordering */
/* Performed on all platforms to make sure that byte order is correct */

#define SCM_HTONL(IP_ADDRESS)         htonl(IP_ADDRESS)
#define SCM_HTONS(PORT_NBR)           htons(PORT_NBR)

#define SCM_NTOHL(IP_ADDRESS)         ntohl(IP_ADDRESS)
#define SCM_NTOHS(PORT_NBR)           ntohs(PORT_NBR)


/* setsockopt */
#ifndef SCM_SETSOCKOPT
  #define SCM_SETSOCKOPT(SOCKET_DESCR, PROT_LEVEL, OPTION, OPT_VAL, OPT_VAL_LEN) \
    setsockopt(SOCKET_DESCR, PROT_LEVEL, OPTION, OPT_VAL, OPT_VAL_LEN)
#endif


#endif






